from functools import wraps, partial
from os import PathLike
from types import TracebackType
from typing import (
    Any, Iterable, Generator, Union, List,
    Sequence, Generic, Hashable,
    Type, TypeVar, Callable, Optional,
)
import threading
from pathlib import Path

from .mapping import Result, Mapper, MapperFunc, compile_mapper_func
from .doublewrap import doublewrap
from .pool import ConnectionPool
from .types import Cursor, CursorParams, Row
from .transaction import Transaction
from .scoped_connection import ScopedConnection

from . import dynamic, static


QueryParams = Union[CursorParams, Any]


class Engine:

    def __init__(
        self,
        templates_paths: Union[str, PathLike, Sequence[Union[str, PathLike]]],
        pool: ConnectionPool,
        mapper: Optional[Mapper] = None,
        commit_on_exit: bool = True,
        str_templates_static_by_default: bool = False,
        identifier_quote_char: str = "'",
    ):
        self.pool = pool
        self.conn = ScopedConnection(pool, commit_on_exit)
        self.mapper = mapper
        if isinstance(templates_paths, str):
            self.templates_paths = [templates_paths]
        elif isinstance(templates_paths, Path):
            self.templates_paths = [str(templates_paths)]
        elif isinstance(templates_paths, Sequence):
            self.templates_paths = templates_paths
        else:
            raise ValueError(
                'templates_paths not an str, '
                'PathLike or Sequence[Str | PathLike]'
            )
        self.dynamic_templates = dynamic.DynamicQueriesCache(
            templates_paths=self.templates_paths,
            identifier_quote_char=identifier_quote_char,
        )
        self.static_templates = static.StaticQueriesCache(
            templates_paths=self.templates_paths,
        )
        self.mapper_cache = {}
        self.mapper_cache_lock = threading.Lock()
        self.str_templates_static_by_default = str_templates_static_by_default

    def get_mapper_from_cache(self, key: Hashable):
        return self.mapper_cache.get(key)

    def cache_mapper(self, key: Hashable, value: MapperFunc):
        with self.mapper_cache_lock:
            self.mapper_cache[key] = value

    def query_from(self, filename: str) -> 'Query':
        if filename.endswith('.sql'):
            create_lazy = self.static_templates.create_lazy
        elif filename.endswith('.sql.tmpl'):
            create_lazy = self.dynamic_templates.create_lazy
        else:
            raise ValueError(f'Unsupported filename extension: {filename}')
        return Query(self, create_lazy(filename=filename))

    def query(self, content: str, static: bool = None) -> 'Query':
        if static is None:
            static = self.str_templates_static_by_default

        if static is True:
            create_lazy = self.static_templates.create_lazy
        elif static is False:
            create_lazy = self.dynamic_templates.create_lazy
        else:
            raise ValueError(f'Unknown "static" arg value: {static}')

        return Query(self, create_lazy(content=content))

    @property
    def cursor(self):
        try:
            return self.conn.cursor()
        except AttributeError:
            raise AttributeError('''
                Trying to access cursor, while not in started state.
                Maybe, you forgot to enter in engine ctx?:
                >>> with engine:
                ...     query.execute(...)
            ''')

    def transaction(self):
        return Transaction(self.conn.__wrapped__)

    def __enter__(self):
        self.conn.__enter__()
        return self

    def __exit__(
            self,
            type_: Optional[Type[BaseException]],
            value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return self.conn.__exit__(type_, value, traceback)

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()


class Query:

    def __init__(
        self,
        engine: Engine,
        lazy_query,
    ):
        self.engine = engine
        self._lazy_query = lazy_query

    def map_to(
        self,
        result: Result = None,
        prefix: Optional[str] = None,
        *,
        mapper: Optional[Mapper] = None,
    ) -> 'MappedQuery[Result]':
        if prefix is None:
            if result is None:
                raise ValueError('Prefix or result must be specified')
            prefix_ = result.__name__.lower()
        else:
            prefix_ = prefix.lower()
        return MappedQuery[Result](
            engine=self.engine,
            lazy_query=self._lazy_query,
            result=prefix_,
            mapper=mapper or self.engine.mapper,
        )

    def execute(
        self,
        params: QueryParams = None,
        /,
        cursor: Cursor = None,
        **kwargs: Any,
    ) -> Cursor:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        return self._lazy_query().execute(
            params or kwargs,
            cursor or self.engine.cursor,
        )

    def executemany(
        self,
        params: Iterable[QueryParams],
        cursor: Cursor = None,
    ) -> Cursor:
        return self._lazy_query().executemany([
            param
            if isinstance(param, (dict, tuple))
            else param.__dict__
            for param in params
        ], cursor or self.engine.cursor,)

    def all(
        self,
        params: QueryParams = None,
        /,
        cursor: Cursor = None,
        **kwargs: Any,
    ):
        if params is None:
            params = kwargs
        else:
            if not isinstance(params, (dict, tuple)):
                params = params.__dict__

        cursor = self._lazy_query().execute(
            params,
            cursor or self.engine.cursor,
        )
        return cursor.fetchall()

    def iter(
        self,
        params: QueryParams = None,
        /,
        batch_: int = 500,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Generator[Any, None, None]:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        cursor_ = self._lazy_query().execute(
            params or kwargs,
            cursor_ or self.engine.cursor,
        )
        while True:
            batch = cursor_.fetchmany(batch_)
            if not batch:
                return
            for row in batch:
                yield row

    def one(
        self,
        params: QueryParams = None,
        /,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Any:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        cursor_ = self._lazy_query().execute(
            params or kwargs,
            cursor_ or self.engine.cursor,
        )
        return cursor_.fetchone()

    def scalar(
        self,
        params: QueryParams = None,
        /,
        raising_: bool = False,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Any:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        row = self.one(
            params or kwargs,
            raising_=raising_,
            cursor_=cursor_ or self.engine.cursor,
        )
        if not raising_ and row is None:
            return None
        return row[0]

    def scalars(
        self,
        params: QueryParams = None,
        /,
        raising_: bool = False,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Generator[Any, None, None]:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        return (
            row[0] for row in self.iter(
                params or kwargs,
                raising_=raising_,
                cursor_=cursor_ or self.engine.cursor,
            )
        )

    def rowcount(
        self,
        params: QueryParams = None,
        /,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> int:
        """Количество строк, обработанных запросом"""

        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        cursor = self._lazy_query().execute(
            params or kwargs,
            cursor_ or self.engine.cursor,
        )
        return cursor.rowcount


class MappedQuery(Generic[Result]):

    def __init__(
        self,
        engine: Engine,
        lazy_query,
        result: str,
        mapper: Mapper,
    ) -> None:
        self.engine = engine
        self._lazy_query = lazy_query
        self.result = result.lower()
        self.mapper = mapper
        self._mapper = None

        # Alias for test simplicity
        self._compile_mapper = compile_mapper_func

    def mapper_func(self, cursor: Cursor) -> Callable[
        [Iterable[Row]],
        Generator[Any, Any, None]
    ]:
        columns = tuple(column[0] for column in cursor.description)
        key = (self.result, self.mapper, columns)
        mapper = self.engine.get_mapper_from_cache(key)
        if not mapper:
            mapper = self._compile_mapper(
                self.result, self.mapper, columns,
            )
            self.engine.cache_mapper(key, mapper)
        return mapper

    def sources(
        self,
        params: QueryParams = None,
        /,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> str:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        cursor = self._lazy_query().execute(
            params or kwargs,
            cursor_ or self.engine.cursor,
        )
        func = self.mapper_func(cursor)
        return getattr(func, 'sources', lambda: '')()

    def all(
        self,
        params: QueryParams = None,
        /,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> List[Result]:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        return list(self.iter(params or kwargs, cursor_=cursor_))

    def iter(
        self,
        params: QueryParams = None,
        /,
        batch: Optional[int] = 500,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Generator[Result, None, None]:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        cursor_ = self._lazy_query().execute(
            params or kwargs,
            cursor_ or self.engine.cursor,
        )
        mapper = self.mapper_func(cursor_)

        if batch:
            fetch = partial(cursor_.fetchmany, batch)
        else:
            fetch = cursor_.fetchall

        def rows_iter():
            while True:
                rows = fetch()
                if not rows:
                    return
                for row in rows:
                    yield row

        for obj in mapper(rows_iter()):
            yield obj

    def one(
        self,
        params: QueryParams = None,
        /,
        batch: int = 500,
        cursor_: Cursor = None,
        **kwargs: Any,
    ) -> Result:
        if params is not None and not isinstance(params, (dict, tuple)):
            params = params.__dict__

        iterator = self.iter(
            params or kwargs,
            batch,
            cursor_ or self.engine.cursor,
        )
        try:
            result = next(iterator)
        except StopIteration:
            iterator.close()
            result = None
        return result


T = TypeVar('T')

@doublewrap
def in_transaction(fn: T, prop: str = 'db') -> T:

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        with getattr(self, prop).transaction():
            return fn(self, *args, **kwargs)

    return wrapper



from collections.abc import Iterable
import threading
from typing import Dict

from markupsafe import Markup
from jinja2 import Template


class Renderer(threading.local):
    bind_params: Dict[str, object]
    param_style: str
    param_index: int

    def _bind_param(self, already_bound, key, value):
        self.param_index += 1
        # TODO: придумать, как с этим жить
        #new_key = '%s_%s' % (key, self.param_index)
        new_key = key
        already_bound[new_key] = value

        param_style = self.param_style
        if param_style == 'qmark':
            return '?'
        elif param_style == 'format':
            return '%s'
        elif param_style == 'numeric':
            return ':%s' % self.param_index
        elif param_style == 'named':
            return ':%s' % new_key
        elif param_style == 'pyformat':
            return '%%(%s)s' % new_key
        elif param_style == 'asyncpg':
            return '$%s' % self.param_index
        else:
            raise AssertionError('Invalid param_style - %s' % param_style)

    def sql_safe(self, value):
        """Filter to mark the value of an expression as safe for inserting
        in a SQL statement"""
        return Markup(value)

    def bind(self, value, name):
        """A filter that prints %s, and stores the value
        in an array, so that it can be bound using a prepared statement

        This filter is automatically applied to every {{variable}}
        during the lexing stage, so developers can't forget to bind
        """
        if isinstance(value, Markup):
            return value
        else:
            return self._bind_param(self.bind_params, name, value)

    def bind_in_clause(self, value):
        values = list(value)
        results = []
        for idx, v in enumerate(values):
            results.append(
                self._bind_param(self.bind_params, f'inclause_{idx}', v)
            )

        clause = ','.join(results)
        clause = '(' + clause + ')'
        return clause

    def build_escape_identifier_filter(self, quote_char):
        def quote_and_escape(value):
            # Escape double quote with 2 double quotes,
            # or escape backtick with 2 backticks
            return (f'{quote_char}'
                    f'{value.replace(quote_char, quote_char * 2)}'
                    f'{quote_char}')

        def identifier_filter(raw_identifier):
            if isinstance(raw_identifier, str):
                raw_identifier = (raw_identifier, )
            if not isinstance(raw_identifier, Iterable):
                raise ValueError(
                    'identifier filter expects a string or an Iterable'
                )
            return Markup('.'.join(
                quote_and_escape(s) for s in raw_identifier
            ))

        return identifier_filter

    def prepare_query(
        self,
        template: Template,
        data: Dict[str, object],
        param_style: str,
    ):
        self.bind_params = {}
        self.param_style = param_style
        self.param_index = 0
        try:
            query = template.render(data)
            if self.param_style in ('named', 'pyformat'):
                bind_params = dict(self.bind_params)
            elif self.param_style in ('qmark', 'numeric', 'format', 'asyncpg'):
                bind_params = list(self.bind_params.values())
            else:
                raise NotImplemented
            return query, bind_params
        finally:
            del self.bind_params
            del self.param_style
            del self.param_index



from dataclasses import dataclass, field
import inspect
from typing import (
    Any, Type, Callable, Tuple, Dict,
    Union, get_args, get_origin, List, Set,
)

from frozendict import frozendict


Factory = Union[Type[Any], Callable[[Any, ...], Any]]


@dataclass(frozen=True, init=False)
class Relationship:
    target: str

    def __init__(self, target: str):
        object.__setattr__(self, 'target', target.lower())


@dataclass(frozen=True, init=False)
class Assign(Relationship):
    pass


@dataclass(frozen=True, init=False)
class Append(Relationship):
    pass


@dataclass(frozen=True, init=False)
class Add(Relationship):
    pass


@dataclass(frozen=True)
class Parameter:
    factory: Factory

    def _parse_relationships(self, rels: Dict[str, Relationship]) -> None:
        try:
            signature = inspect.signature(self.factory)
        except ValueError:
            object.__setattr__(self, 'relationships', frozendict(rels))
            return

        new_rels = {}
        for name_, sign_param in signature.parameters.items():
            # User parameters have priority
            if rel := rels.get(name_):
                new_rels[name_] = rel
                continue

            # Trying to create relationship automatically
            annot = sign_param.annotation
            origin = get_origin(annot)
            args = get_args(annot)
            if origin is None:
                new_rels[name_] = Assign(annot.__name__.lower())
            elif issubclass(origin, List):
                new_rels[name_] = Append(args[0].__name__.lower())
            elif issubclass(origin, Set):
                new_rels[name_] = Add(args[0].__name__.lower())

        object.__setattr__(self, 'relationships', frozendict(new_rels))


@dataclass(frozen=True, init=False)
class Entity(Parameter):
    id: Union[str, Tuple[str, ...]]
    relationships: Dict[str, Relationship] = field(default_factory=frozendict)

    def __init__(
        self,
        factory: Factory,
        id_: Union[str, Tuple[str, ...]],
        /,
        **relationships: Relationship,
    ) -> None:
        # Set cls
        object.__setattr__(self, 'factory', factory)

        # Set ID
        object.__setattr__(
            self, 'id',
            (id_.lower(),)
            if isinstance(id_, str) else
            tuple((_.lower() for _ in id_))
        )

        # Set relationships
        self._parse_relationships(relationships)


@dataclass(frozen=True, init=False)
class Value(Parameter):
    reduce_none: bool
    relationships: Dict[str, Relationship] = field(default_factory=frozendict)

    def __init__(
        self,
        factory: Factory,
        reduce_none: bool = True,
        /,
        **relationships: Relationship,
    ) -> None:
        # Set cls
        object.__setattr__(self, 'factory', factory)

        # Set ID
        object.__setattr__(self, 'reduce_none', reduce_none)

        # Set relationships
        self._parse_relationships(relationships)


@dataclass(frozen=True, init=False)
class Mapper:
    params: frozendict[str, Parameter] = field(default_factory=frozendict)

    def __init__(self, **params: Parameter) -> None:
        object.__setattr__(self, 'params', frozendict((
            (key.lower(), value)
            for key, value in params.items()
        )))



from classic.db_tools import Engine, ConnectionPool

from .conftest import SQL_DIR_PATH, create_pool


def test_engine_commits(conn_pool: ConnectionPool) -> None:
    engine = Engine(SQL_DIR_PATH, conn_pool)
    with engine:
        engine.query(
            'CREATE TABLE example(a int, b int)'
        ).execute()
        engine.query(
            'INSERT INTO example(a, b) '
            'VALUES (1, 2), (2, 3), (3, 4)'
        ).execute()

    with engine:
        assert engine.query(
            'SELECT * FROM example'
        ).all() == [(1, 2), (2, 3), (3, 4)]

    with engine:
        engine.query(
            'DROP TABLE example'
        ).execute()


def test_engine_with_autocommit() -> None:
    conn_pool = create_pool(dict(autocommit=True), dict(limit=1))
    engine = Engine(SQL_DIR_PATH, conn_pool, commit_on_exit=False)
    with engine:
        engine.query(
            'CREATE TABLE example(a int, b int)'
        ).execute()
        engine.query(
            'INSERT INTO example(a, b) '
            'VALUES (1, 2), (2, 3), (3, 4)'
        ).execute()

    with engine:
        assert engine.query(
            'SELECT * FROM example'
        ).all() == [(1, 2), (2, 3), (3, 4)]

    with engine:
        engine.query('DROP TABLE example').execute()


def test_engine_with_autocommit_and_tx() -> None:
    conn_pool = create_pool(dict(autocommit=True), dict(limit=1))
    engine = Engine(SQL_DIR_PATH, conn_pool, commit_on_exit=False)
    with engine:
        with engine.transaction():
            engine.query(
                'CREATE TABLE example(a int, b int)'
            ).execute()
            engine.query(
                'INSERT INTO example(a, b) '
                'VALUES (1, 2), (2, 3), (3, 4)'
            ).execute()

    with engine:
        assert engine.query(
            'SELECT * FROM example'
        ).all() == [(1, 2), (2, 3), (3, 4)]

    with engine:
        with engine.transaction():
            engine.query(
                'DROP TABLE example'
            ).execute()



import threading
from types import TracebackType
from typing import Any, Optional, Type

from .pool import ConnectionPool
from .types import Connection


class ScopedConnection(threading.local):
    _conn_pool: ConnectionPool

    def __init__(
            self,
            conn_pool: ConnectionPool,
            commit_on_exit: bool = True,
    ):
        super().__init__()
        self._conn_pool = conn_pool
        self._commit_on_exit = commit_on_exit
        self._started = False

    def __enter__(self) -> Connection:
        self._conn = self._conn_pool.getconn()
        self._started = True
        return self._conn

    def __exit__(
            self,
            type_: Optional[Type[BaseException]],
            value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        if not hasattr(self, '_conn'):
            return False

        if self._conn.autocommit is False:
            if type_ is None and self._commit_on_exit:
                self._conn.commit()
            else:
                self._conn.rollback()

        self._conn_pool.release(self._conn)
        del self._conn
        return False

    def __getattr__(self, item: str) -> Any:
        if not self._started:
            raise AttributeError(f'''
                Trying to access {item}, while not in started state.
                Maybe, you forgot to enter in ScopedConnection?:
                >>> with ScopedConnection(pool) as conn:
                ...     query.execute(conn)
            ''')
        return getattr(self._conn, item)

    @property
    def __wrapped__(self) -> Connection:
        return self._conn



from jinja2.ext import Extension
from jinja2.lexer import Token


class AutoBind(Extension):

    def extract_param_name(self, tokens):
        name = ''
        for token in tokens:
            if token.test('variable_begin'):
                continue
            elif token.test('name'):
                name += token.value
            elif token.test('dot'):
                name += token.value
            else:
                break
        if not name:
            name = 'bind#0'
        return name

    def filter_stream(self, stream):
        """
        We convert
        {{ some.variable | filter1 | filter 2}}
            to
        {{ ( some.variable | filter1 | filter 2 ) | bind}}

        ... for all variable declarations in the template

        Note the extra ( and ). We want the | bind to apply to the entire value,
        not just the last value.
        The parentheses are mostly redundant, except in expressions like
        {{ '%' ~ myval ~ '%' }}

        This function is called by jinja2 immediately
        after the lexing stage, but before the parser is called.
        """
        while not stream.eos:
            token = next(stream)
            if token.test('variable_begin'):
                var_expr = []
                while not token.test('variable_end'):
                    var_expr.append(token)
                    token = next(stream)
                variable_end = token

                last_token = var_expr[-1]
                lineno = last_token.lineno

                # don't bind twice
                if (
                        not last_token.test('name') or
                        last_token.value not in ('bind', 'inclause', 'sqlsafe')
                ):
                    param_name = self.extract_param_name(var_expr)

                    var_expr.insert(1, Token(lineno, 'lparen', '('))
                    var_expr.append(Token(lineno, 'rparen', ')'))
                    var_expr.append(Token(lineno, 'pipe', '|'))
                    var_expr.append(Token(lineno, 'name', 'bind'))
                    var_expr.append(Token(lineno, 'lparen', '('))
                    var_expr.append(Token(lineno, 'string', param_name))
                    var_expr.append(Token(lineno, 'rparen', ')'))

                var_expr.append(variable_end)
                for token in var_expr:
                    yield token
            else:
                yield token



from unittest.mock import Mock, call, MagicMock

from classic.db_tools import Engine
from classic.db_tools.types import Cursor


def test_queries_cache(engine: Engine):
    for query in (
        engine.query_from('test_render.sql'),
        engine.query_from('test_render.sql.tmpl'),
        engine.query('SELECT 1'),
        engine.query('SELECT 1', static=True),
    ):
        execute = Mock()
        lazy_query = MagicMock()
        lazy_query.__call__ = Mock(return_value=execute)
        cursor = Mock(Cursor)

        query._lazy_query = lazy_query
        query.execute({}, cursor=cursor)
        query.execute({}, cursor=cursor)

        lazy_query.assert_has_calls([
            call(),
            call().execute({}, cursor),
            call(),
            call().execute({}, cursor),
        ])



from classic.db_tools import ConnectionPool

import psycopg


def test_connection_pool(conn_pool: ConnectionPool):
    with conn_pool.connect() as connection:
        assert isinstance(connection, psycopg.Connection)



from dataclasses import dataclass

import pytest

from classic.db_tools import Engine, in_transaction
import psycopg


@dataclass
class SomeCls:
    db: Engine

    @in_transaction
    def run(self):
        self.db.query("INSERT INTO tasks VALUES ('1', '2')")
        raise ValueError


def test__in_transaction(engine: Engine, ddl):
    with pytest.raises(ValueError):
        SomeCls(db=engine).run()

    with pytest.raises(psycopg.errors.UndefinedTable):
        engine.query('SELECT * FROM tasks').execute()



from dataclasses import dataclass

import pytest

from classic.db_tools import Engine

from .dto import Task


@dataclass
class Task:
    name: str
    value: str


@pytest.fixture
def tasks(engine: Engine, ddl):
    engine.query_from('example/save_task.sql').execute(
        Task(name='1', value='value_1'),
    )
    engine.query_from('example/save_task.sql').executemany([
        {'name': '2', 'value': 'value_2'},
        {'name': '3', 'value': 'value_3'},
    ])
    engine.query_from('example/save_task_statuses.sql').executemany([
        {'title': 'ready', 'task_id': 1},
        {'title': 'active', 'task_id': 2},
        {'title': 'completed', 'task_id': 3},
    ])
    yield


def test_iter(engine: Engine, tasks):
    result = engine.query_from('example/find_by_name.sql.tmpl').iter(
        dict(name='1'),
    )
    assert next(result) == (1, '1')
    with pytest.raises(StopIteration):
        next(result)


def test_all(engine: Engine, tasks):
    assert engine.query_from(
        'example/find_by_name.sql.tmpl'
    ).all(name='1') == [(1, '1')]


def test_one(engine: Engine, tasks):
    assert engine.query_from(
        'example/get_by_id.sql'
    ).one(id='1') == (1, '1')


def test_scalar(engine: Engine, tasks):
    assert engine.query_from(
        'example/get_by_id.sql'
    ).scalar(id='1') == 1


def test_scalars(engine: Engine, tasks):
    assert list(
        engine.query_from(
            'example/get_all.sql'
        ).scalars()
    ) == [1, 2, 3]


def test_insert(engine: Engine, ddl):
    assert engine.query_from('example/count.sql').scalar() == 0

    row_id = engine.query_from('example/save_task.sql').scalar(
        Task(name='1', value='value_1'),
    )

    assert (
        engine.query_from('example/get_all.sql').one(id=row_id)
        == (row_id, '1', 'value_1')
    )


def test_insert_many(engine: Engine, ddl):
    assert engine.query_from('example/count.sql').scalar() == 0

    engine.query_from('example/save_task.sql').executemany([
        Task(name='1', value='value_1'),
        {'name': '2', 'value': 'value_2'},
        {'name': '3', 'value': 'value_3'},
    ])

    assert engine.query_from('example/get_all.sql').all() == [
        (1, '1', 'value_1'),
        (2, '2', 'value_2'),
        (3, '3', 'value_3'),
    ]


@pytest.mark.parametrize(
    'value,title', [
        (1, 'ready'),
        (2, 'active'),
        (3, 'completed'),
    ]
)
def test_get_by_status(engine: Engine, tasks, title, value):
    assert engine.query_from(
        'example/joined_get_by_status.sql.tmpl'
    ).one(title=title) == (value, str(value), title)


def test_get_by_status_none(engine: Engine, tasks):
    assert engine.query_from(
        'example/joined_get_by_status.sql.tmpl').all() == [
        (1, '1'), (2, '2'), (3, '3'),
    ]


@pytest.mark.parametrize(
    'value,title', [
        (1, 'ready'),
        (1, 'active'),
        (1, 'completed'),
        (3, None)
    ]
)
def test_count_by_status(engine: Engine, tasks, value, title):
    assert engine.query_from(
        'example/count_by_status.sql.tmpl'
    ).scalar(title=title) == value


@pytest.mark.parametrize(
    'value,title', [
        (1, 'ready'),
        (2, 'active'),
        (3, 'completed'),
    ]
)
def test_sum_tasks(engine: Engine, tasks, title, value):
    assert engine.query_from(
        'example/sum_tasks.sql.tmpl'
    ).scalar(title=title) == value



import sys
from typing import Union

from classic.db_tools.types import Connection, Cursor

known_styles = {}


def recognize_param_style(obj: Union[Connection, Cursor]) -> str:
    cls = obj.__class__

    try:
        return known_styles[cls]
    except KeyError:
        modname = cls.__module__
        while modname:
            try:
                style = sys.modules[modname].paramstyle  # type: ignore
                known_styles[cls] = style
                return style
            except AttributeError:
                if "." in modname:
                    modname = modname.rsplit(".", 1)[0]
                else:
                    break
    raise TypeError(f"Can't find paramstyle for connection {obj!r}")



import ast
from collections import defaultdict
from itertools import chain
from typing import (
    Iterable, List, Tuple, Dict, Optional,
    Generator, Callable, TypeVar, Literal,
)

from .params import (
    Parameter, Relationship, Assign, Append, Entity, Value, Mapper,
)
from ..types import Row


Accessor = Literal['attr', 'item']
Result = TypeVar('Result')
MapperFunc = Callable[[], Generator[Result, Row, None]]

# Aliases for readability
Prefix = str
Field = str
Column = str

# Tuple positions for fields_to_columns
index = 0
col = 1


class Context:
    mapper: Dict[Prefix, Parameter]
    rels: Dict[Prefix, Dict[Field, Relationship]]
    result: Tuple[Prefix, Parameter]
    columns: Optional[Tuple[Column, ...]]
    fields_to_columns: Dict[
        Parameter, Dict[Field, Tuple[int, Column]]
    ]

    def __init__(
        self,
        result: str,
        mapper: Mapper,
        columns: Tuple[Column, ...],
    ):
        self.mapper = {}
        self.rels = defaultdict(dict)
        self.result = (result, mapper.params[result])
        self.columns = None
        self.fields_to_columns = defaultdict(dict)
        self.parse_columns(columns, mapper)

    @staticmethod
    def accessor(param: Parameter) -> Accessor:
        return 'item' if issubclass(param.factory, dict) else 'attr'

    def namespace(self):
        return {
            mapper.factory.__name__: mapper.factory
            for mapper in self.mapper.values()
        }

    def column_for_field(self, param: Parameter, field: Field) -> str:
        try:
            return self.fields_to_columns[param][field][col]
        except KeyError:
            raise ValueError(f'For class {param} not found field {field}')

    def parse_columns(
        self,
        columns: Tuple[Column, ...],
        mapper: Mapper,
    ):
        self.columns = columns
        for index_, column in enumerate(columns):
            try:
                prefix, field_name = column.lower().split('__')
            except ValueError as e:
                raise ValueError(
                    f'Column {column} are not contains name of cls '
                    f'and name of field, concatenated with __'
                ) from e

            param = mapper.params.get(prefix)
            if not param:
                raise ValueError(f'Prefix {prefix} not found in mapper')

            if param not in self.mapper:
                self.mapper[prefix] = param

                for prefix_, param_ in mapper.params.items():
                    for rel_field, rel in param_.relationships.items():
                        if (
                            rel.target == prefix and
                            rel not in self.rels[prefix]
                        ):
                            self.rels[prefix_][rel_field] = rel

            self.fields_to_columns[param][field_name] = index_, column


def compile_mapper_func(
    result: Result,
    mapper: Mapper,
    columns: Tuple[Column, ...],
) -> MapperFunc:
    ctx = Context(result, mapper, columns)

    # Mapping class_name: Class
    namespace = ctx.namespace()

    # AST Tree with module
    ast_module = render_module(ctx)

    # Compile func
    code = compile(ast_module, '<string>', 'exec')
    exec(code, namespace)

    func = namespace['mapper_func']

    # For easy debugging add source code to func
    func.sources = lambda: ast.unparse(ast_module)

    return func


def render_identity_maps(ctx: Context) -> Iterable[ast.stmt]:
    for prefix in ctx.mapper:
        yield ast.Assign(
            [ast.Name(f'{prefix}_map', ast.Store())],
            ast.Dict(keys=[], values=[]),
        )


def render_last_root() -> ast.stmt:
    return ast.Assign(
        [ast.Name("last_root", ast.Store())],
        ast.Constant(None),
    )


def render_cycle(ctx: Context) -> ast.stmt:
    return ast.For(
        target=ast.Name("row", ast.Store()),
        iter=ast.Name("rows", ast.Load()),
        body=list(render_cycle_body(ctx)),
        orelse=[],
    )


def render_check_for_none(columns: List[int]) -> ast.stmt:
    if len(columns) == 1:
        stmt = ast.Compare(
            left=ast.Subscript(
                ast.Name("row", ast.Load()),
                ast.Constant(columns[0]),
                ast.Load(),
            ),
            ops=[ast.Is()],
            comparators=[ast.Constant(None)],
        )
    else:
        stmt = ast.BoolOp(
            op=ast.And(),
            values=[
                ast.Compare(
                    left=ast.Subscript(
                        ast.Name("row", ast.Load()),
                        ast.Constant(column),
                        ast.Load(),
                    ),
                    ops=[ast.Is()],
                    comparators=[ast.Constant(None)],
                )
                for column in columns
            ],
        )
    return ast.If(
        test=stmt,
        body=[ast.Return(ast.Constant(None))],
        orelse=[],
    )


def render_factory_call(
    ctx: Context, prefix: Prefix, param: Parameter,
) -> ast.Assign:
    return ast.Assign(
        [ast.Name(prefix, ast.Store())],
        ast.Call(
            func=ast.Name(param.factory.__name__, ast.Load()),
            args=[],
            keywords=[
                ast.keyword(
                    field,
                    (
                        ast.Subscript(
                            ast.Name("row", ast.Load()),
                            ast.Constant(column[index]),
                            ast.Load(),
                        )
                    ),
                )
                for field, column in ctx.fields_to_columns[param].items()
            ],
        ),
    )


def render_mapper_call(prefix: Prefix) -> ast.expr:
    return ast.Call(
        func=ast.Name(f'map_{prefix}', ast.Load()),
        args=[ast.Name("row", ast.Load())],
        keywords=[],
    )


def render_rel_factory_call(
    left: Prefix,
    field: Field,
    rel: Relationship,
    accessor: Accessor,
) -> Generator[ast.stmt, None, None]:
    right: Prefix = rel.target

    if isinstance(rel, Assign):
        target = (
            ast.Attribute(
                ast.Name(left, ast.Load()),
                field,
                ast.Store(),
            )
            if accessor == "attr"
            else ast.Subscript(
                ast.Name(left, ast.Load()),
                ast.Constant(field),
                ast.Store(),
            )
        )
        yield ast.Assign([target], render_mapper_call(right))

    elif isinstance(rel, Append):
        if accessor == "item":
            yield ast.If(
                test=ast.Compare(
                    left=ast.Constant(field),
                    ops=[ast.NotIn()],
                    comparators=[ast.Name(left, ast.Load())],
                ),
                body=[
                    ast.Assign(
                        [
                            ast.Subscript(
                                ast.Name(left, ast.Load()),
                                ast.Constant(field),
                                ast.Store(),
                            ),
                        ],
                        ast.List([], ast.Load()),
                    )
                ],
                orelse=[],
            )
            value = ast.Subscript(
                ast.Name(left, ast.Load()),
                ast.Constant(field),
                ast.Load(),
            )
        else:
            value = ast.Attribute(
                ast.Name(left, ast.Load()), field, ast.Load(),
            )

        yield ast.If(
            test=ast.NamedExpr(
                target=ast.Name("obj", ast.Store()),
                value=ast.Call(
                    func=ast.Name(f'map_{right}', ast.Load()),
                    args=[ast.Name("row", ast.Load())],
                    keywords=[],
                ),
            ),
            body=[
                ast.Expr(
                    ast.Call(
                        func=ast.Attribute(value, "append", ast.Load()),
                        args=[render_mapper_call(right)],
                        keywords=[],
                    )
                )
            ],
            orelse=[],
        )
    else:
        raise NotImplemented


def render_obj_id(ctx: Context, prefix: Prefix, entity: Entity) -> ast.stmt:
    if len(entity.id) == 1:
        id_field = entity.id[0]
        id_col = ctx.fields_to_columns[entity][id_field][index]
        id_val = ast.Subscript(
            ast.Name("row", ast.Load()),
            ast.Constant(id_col),
            ast.Load(),
        )
    else:
        id_val = ast.Tuple(
            [
                ast.Subscript(
                    ast.Name("row", ast.Load()),
                    ast.Constant(ctx.fields_to_columns[entity][field][index]),
                    ast.Load(),
                )
                for field in entity.id
            ],
            ast.Load(),
        )

    return ast.Assign([ast.Name(f'{prefix}_id', ast.Store())], id_val)


def render_get_or_create(
    ctx: Context, prefix: Prefix, param: Parameter,
) -> Generator[ast.stmt, None, None]:
    yield ast.Assign(
        [ast.Name(prefix, ast.Store())],
        ast.Call(
            func=ast.Attribute(
                ast.Name(f'{prefix}_map', ast.Load()), "get", ast.Load()
            ),
            args=[ast.Name(f'{prefix}_id', ast.Load())],
            keywords=[],
        ),
    )

    assign = render_factory_call(ctx, prefix, param)
    assign.targets.append(
        ast.Subscript(
            ast.Name(f'{prefix}_map', ast.Load()),
            ast.Name(f'{prefix}_id', ast.Load()),
            ast.Store(),
        ),
    )
    yield ast.If(
        test=ast.Compare(
            left=ast.Name(prefix, ast.Load()),
            ops=[ast.Is()],
            comparators=[ast.Constant(None)],
        ),
        body=[assign],
        orelse=[],
    )


def render_entity(ctx: Context, prefix: Prefix, entity: Entity) -> ast.stmt:
    id_columns = [
        cols[index]
        for field, cols in ctx.fields_to_columns[entity].items()
        if field in entity.id
    ]
    body = [
        render_check_for_none(id_columns),
        render_obj_id(ctx, prefix, entity),
        *render_get_or_create(ctx, prefix, entity),
        *chain.from_iterable(
            render_rel_factory_call(
                prefix,
                field,
                rel,
                ctx.accessor(entity),
            )
            for field, rel in ctx.rels[prefix].items()
        ),
        ast.Return(
            ast.Name(prefix, ast.Load()),
        ),
    ]
    return ast.FunctionDef(
        name=f'map_{prefix}',
        args=ast.arguments(
            posonlyargs=[],
            args=[
                ast.arg(arg="rows"),
            ],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=body,
        decorator_list=[],
    )


def render_value(
    ctx: Context,
    prefix: Prefix,
    value: Value,
) -> ast.stmt:
    columns = [col_[index] for col_ in ctx.fields_to_columns[value].values()]
    if_body = []
    if value.reduce_none:
        if_body.append(render_check_for_none(columns))
    if_body += [
        render_factory_call(ctx, prefix, value),
        *chain.from_iterable(
            render_rel_factory_call(
                prefix,
                field,
                rel,
                ctx.accessor(value),
            )
            for field, rel in ctx.rels[prefix].items()
        ),
        ast.Return(
            ast.Name(prefix, ast.Load()),
        ),
    ]
    return ast.FunctionDef(
        name=f'map_{prefix}',
        args=ast.arguments(
            posonlyargs=[],
            args=[
                ast.arg(arg="rows"),
            ],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=if_body,
        decorator_list=[],
    )


def render_mappers(ctx: Context) -> Generator[ast.stmt, None, None]:
    for prefix, mapper in ctx.mapper.items():
        if isinstance(mapper, Entity):
            yield render_entity(ctx, prefix, mapper)
        elif isinstance(mapper, Value):
            yield render_value(ctx, prefix, mapper)
        else:
            raise NotImplemented


def render_cycle_body(ctx: Context) -> Generator[ast.stmt, None, None]:
    yield ast.Assign(
        [ast.Name("root", ast.Store())],
        render_mapper_call(ctx.result[0]),
    )
    yield ast.If(
        test=ast.Compare(
            left=ast.Name("last_root", ast.Load()),
            ops=[ast.IsNot()],
            comparators=[ast.Name("root", ast.Load())],
        ),
        body=[
            ast.If(
                test=ast.Compare(
                    left=ast.Name("last_root", ast.Load()),
                    ops=[ast.IsNot()],
                    comparators=[ast.Constant(None)],
                ),
                body=[ast.Expr(ast.Yield(ast.Name("last_root", ast.Load())))],
                orelse=[],
            ),
            ast.Assign(
                targets=[ast.Name("last_root", ast.Store())],
                value=ast.Name("root", ast.Load()),
            ),
        ],
        orelse=[],
    )


def render_post_cycle() -> ast.stmt:
    return ast.If(
        test=ast.Compare(
            left=ast.Name("last_root", ast.Load()),
            ops=[ast.IsNot()],
            comparators=[ast.Constant(None)],
        ),
        body=[ast.Expr(ast.Yield(ast.Name("last_root", ast.Load())))],
        orelse=[],
    )


def render_mapper_func(ctx: Context) -> ast.stmt:
    return ast.FunctionDef(
        name="mapper_func",
        args=ast.arguments(
            posonlyargs=[],
            args=[
                ast.arg(arg="rows"),
            ],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=[
            *render_identity_maps(ctx),
            *render_mappers(ctx),
            render_last_root(),
            render_cycle(ctx),
            render_post_cycle(),
        ],
        decorator_list=[],
    )


def render_module(ctx: Context) -> ast.Module:
    func = render_mapper_func(ctx)
    module = ast.Module(body=[func], type_ignores=[])
    return ast.fix_missing_locations(module)



from functools import wraps


def doublewrap(fn):
    """
    Классный сниппет, облегчающий создание декораторов с параметрами.
    Взято отсюда: https://stackoverflow.com/a/14412901
    В сравнении с оригиналом чуть-чуть улучшена читаемость.

    Пример:

    >>> @doublewrap
    ... def with_default(method, default):
    ...
    ...     @wraps(method)
    ...     def wrapper(self, *args, **kwargs):
    ...         return method(self, *args, **kwargs) or default
    ...
    ...     return wrapper
    ...
    ... @with_default(1)
    ... def return_none():
    ...     return None
    ...
    ... return_none()
    1
    """

    @wraps(fn)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return fn(args[0])
        else:
            # decorator arguments
            return lambda real_fn: fn(real_fn, *args, **kwargs)

    return new_dec



from classic.db_tools import Engine

import pytest

from .conftest import SQL_DIR_PATH


def test_engine_works_with_outer_cursor(conn_pool) -> None:
    engine = Engine(SQL_DIR_PATH, conn_pool, commit_on_exit=False)

    with conn_pool.connect() as conn:
        assert engine.query('SELECT 1').scalar(cursor_=conn.cursor()) == 1


def test_engine_fail_without_enter_cx(conn_pool):
    engine = Engine(SQL_DIR_PATH, conn_pool, commit_on_exit=False)

    with pytest.raises(AttributeError):
        assert engine.query('SELECT 1').scalar() == 1



import threading
from typing import Iterable, Callable, Sequence

import jinja2

from classic.db_tools.params_styles import recognize_param_style
from classic.db_tools.types import Cursor, CursorParams

from .renderer import Renderer
from .extension import AutoBind


class DynamicQuery:

    def __init__(
        self,
        renderer: Renderer,
        template: jinja2.Template,
    ):
        self.renderer = renderer
        self.template = template

    def execute(
        self,
        params: CursorParams = None,
        cursor: Cursor = None,
    ) -> Cursor:
        sql, ordered_params = self.renderer.prepare_query(
            self.template, params, recognize_param_style(cursor),
        )
        cursor.execute(sql, ordered_params)
        return cursor

    def executemany(
        self,
        params: Iterable[CursorParams],
        cursor: Cursor,
    ) -> Cursor:
        for param in params:
            sql, ordered_params = self.renderer.prepare_query(
                self.template, param, recognize_param_style(cursor),
            )
            cursor.execute(sql, ordered_params)
        return cursor


class DynamicQueriesCache:
    VALID_ID_QUOTE_CHARS = ('`', "'")

    def __init__(
        self,
        templates_paths: Sequence[str],
        identifier_quote_char: str = "'",
    ):
        assert identifier_quote_char in self.VALID_ID_QUOTE_CHARS

        self.identifier_quote_char = identifier_quote_char
        self.jinja = jinja2.Environment(
            loader=jinja2.FileSystemLoader(templates_paths),
            auto_reload=False,
            autoescape=True,
        )
        self.renderer = Renderer()
        self.jinja.add_extension(AutoBind)
        self.jinja.filters['bind'] = self.renderer.bind
        self.jinja.filters['sqlsafe'] = self.renderer.sql_safe
        self.jinja.filters['inclause'] = self.renderer.bind_in_clause
        self.jinja.filters['identifier'] = (
            self.renderer.build_escape_identifier_filter(
                self.identifier_quote_char,
            )
        )
        self.cache = {}
        self.lock = threading.RLock()

    def create_lazy(
        self,
        filename: str = None,
        content: str = None,
    ) -> Callable[[], DynamicQuery]:
        if filename:
            key = filename
        elif content:
            key = content
        else:
            raise NotImplemented

        def lazy_query():
            with self.lock:
                obj = self.cache.get(key)
                if obj is None:
                    if filename:
                        template = self.jinja.get_template(filename)
                    elif content:
                        template = self.jinja.from_string(content)
                    else:
                        raise NotImplemented

                    obj = DynamicQuery(self.renderer, template)
                    self.cache[key] = obj

            return obj

        return lazy_query



from unittest.mock import Mock

from classic.db_tools import Engine

from .dto import Task


def fake(rows):
    while True:
        yield


def test_queries_cache(engine: Engine):
    fake_compile = Mock(return_value=fake)
    query = engine.query('SELECT 1 WHERE FALSE').map_to(Task)
    query._compile_mapper = fake_compile

    fake_compile.assert_not_called()

    query.one()
    query.one()

    fake_compile.assert_called_once()



from dataclasses import dataclass

import pytest

from classic.db_tools import Engine, Mapper, Entity, Append, Value

from .dto import Task, Status


sql = '''
    SELECT
        tasks.id            AS task__id,
        tasks.name          AS task__name,
        task_status.id      AS status__id,
        task_status.title   AS status__title
    FROM tasks
    LEFT JOIN task_status ON task_status.task_id = tasks.id
    ORDER BY tasks.id, task_status.id 
'''


mapper = Mapper(
    task=Entity(Task, 'id', statuses=Append('status')),
    status=Value(Status),
)


@pytest.fixture
def tasks(engine: Engine, ddl):
    engine.query_from('example/save_task.sql').executemany([
        {'name': 'First', 'value': ''},
        {'name': 'Second', 'value': ''},
        {'name': 'Third', 'value': ''},
        {'name': 'Four', 'value': ''},
    ])
    engine.query_from('example/save_task_statuses.sql').executemany([
        {'title': 'CREATED', 'task_id': 1},
        {'title': 'CREATED', 'task_id': 2},
        {'title': 'CREATED', 'task_id': 3},
        {'title': 'STARTED', 'task_id': 1},
        {'title': 'FINISHED', 'task_id': 1},
    ])
    yield


@pytest.mark.parametrize('static', (True, False))
def test_returning_with_rels__all(engine: Engine, ddl, tasks, static):
    objects = engine.query(sql, static=static).map_to(Task, mapper=mapper).all()
    assert objects == [
        Task(id=1, name='First', statuses=[
            Status(id=1, title='CREATED'),
            Status(id=4, title='STARTED'),
            Status(id=5, title='FINISHED'),
        ]),
        Task(id=2, name='Second', statuses=[
            Status(id=2, title='CREATED'),
        ]),
        Task(id=3, name='Third', statuses=[
            Status(id=3, title='CREATED'),
        ]),
        Task(id=4, name='Four', statuses=[]),
    ]


@pytest.mark.parametrize('static', (True, False))
def test_returning_with_rels__one(engine: Engine, ddl, tasks, static):
    obj = engine.query(sql, static=static).map_to(Task, mapper=mapper).one()
    assert obj == (
        Task(id=1, name='First', statuses=[
            Status(id=1, title='CREATED'),
            Status(id=4, title='STARTED'),
            Status(id=5, title='FINISHED'),
        ])
    )


def test_custom_name(engine: Engine):
    objects = engine.query('''
    SELECT 
        data.task_id        AS custom__id,
        data.task_name      AS custom__name,
        data.status_id      AS another__id,
        data.status_title   AS another__title
    FROM (
        VALUES
            (1, 'First', 1, 'CREATED'),
            (1, 'First', 4, 'STARTED'),
            (1, 'First', 5, 'FINISHED')
    ) AS data(task_id, task_name, status_id, status_title)
    ''').map_to(
        prefix='custom',
        mapper=Mapper(
            custom=Entity(Task, 'id', statuses=Append('another')),
            another=Entity(Status, 'id'),
        ),
    ).one()
    assert objects == Task(id=1, name='First', statuses=[
        Status(id=1, title='CREATED'),
        Status(id=4, title='STARTED'),
        Status(id=5, title='FINISHED'),
    ])


@dataclass
class SomeObj:
    id: int
    value: str


@dataclass
class AnotherObj:
    id: int
    some_obj: SomeObj = None


def test_one_to_one(engine: Engine):
    objects = engine.query('''
    SELECT 
        data.AnotherObj__id as AnotherObj__id,
        data.SomeObj__id as SomeObj__id,
        data.SomeObj__value as SomeObj__value
    FROM (
        VALUES
            (1, 1, 'VALUE'),
            (2, 1, 'VALUE')
    ) AS data(AnotherObj__id, SomeObj__id, SomeObj__value)
    ''').map_to(
        AnotherObj,
        mapper=Mapper(
            AnotherObj=Entity(AnotherObj, 'id'),
            SomeObj=Entity(SomeObj, 'id'),
        )
    ).all()
    assert objects == [
        AnotherObj(id=1, some_obj=SomeObj(1, 'VALUE')),
        AnotherObj(id=2, some_obj=SomeObj(1, 'VALUE')),
    ]



from classic.db_tools import Engine


def test_from_file(engine: Engine):
    assert engine.query_from('test_render.sql').scalar() == 'rendered'


def test_from_str(engine: Engine):
    assert engine.query("SELECT 'rendered'").scalar() == 'rendered'



# from dataclasses import dataclass
# from typing import Annotated
#
# from classic.db_tools import Engine,
#
#
# @dataclass
# class Nested:
#     field_4: int
#
#
# @dataclass
# class SomeObj:
#     field_1: int
#     field_2: int
#     field_3: int
#
#
# def test_composite_keys_mapping(engine: Engine):
#     query = engine.query('''
#         SELECT
#             field_1 as SomeObj__field_1,
#             field_2 as SomeObj__field_2,
#             field_3 as SomeObj__field_3,
#             field_4 as Nested__field_4
#         FROM (
#             VALUES
#                 (1, 1, 1, 1),
#                 (1, 1, 2, 2),
#                 (1, 2, 3, 3),
#                 (1, 3, 4, 4)
#         ) AS data(field_1, field_2, field_3, field_4)
#     ''').map_to(
#         tuple[
#             Annotated[SomeObj, ID('field_1', 'field_2')],
#             Annotated[Nested, ID('field_4')],
#         ],
#     )
#     assert query.all() == [
#         (SomeObj(1, 1, 1), Nested(1)),
#         (SomeObj(1, 1, 1), Nested(2)),
#         (SomeObj(1, 2, 3), Nested(3)),
#         (SomeObj(1, 3, 4), Nested(4)),
#     ]



from typing import Sequence, Callable

import os
import threading

from classic.db_tools.types import Cursor, CursorParams


class StaticQuery:

    def __init__(
        self,
        filepath: str = None,
        content: str = None,
    ):
        assert filepath is None or content is None
        if filepath:
            self.filepath = filepath
            with open(self.filepath, 'rt') as file:
                self.content = file.read()
        elif content:
            self.filepath = None
            self.content = content
        else:
            raise NotImplemented

    def execute(
        self,
        params: CursorParams = None,
        cursor: Cursor = None,
    ) -> Cursor:
        cursor.execute(self.content, params)
        return cursor

    def executemany(
        self,
        params: Sequence[CursorParams],
        cursor: Cursor = None,
    ) -> Cursor:
        cursor.executemany(self.content, params)
        return cursor


class StaticQueriesCache:

    def __init__(self, templates_paths: Sequence[str]):
        self.cache = {}
        self.templates_paths = templates_paths
        self.lock = threading.RLock()

    def create_lazy(
        self,
        filename: str = None,
        content: str = None,
    ) -> Callable[[], StaticQuery]:
        if filename:
            key = filename
        elif content:
            key = content
        else:
            raise NotImplemented

        def lazy_query():
            with self.lock:
                obj = self.cache.get(key)
                if obj is None:
                    if filename:
                        for path in self.templates_paths:
                            filepath = os.path.join(path, filename)
                            if os.path.exists(filepath):
                                obj = StaticQuery(filepath=filepath)
                                break
                        if obj is None:
                            raise FileNotFoundError(
                                f'File {filename} does not exist in '
                                f'{self.templates_paths} dirs'
                            )
                    elif content:
                        obj = StaticQuery(content=content)
                    else:
                        raise NotImplemented

                    with self.lock:
                        self.cache[key] = obj

            return obj

        return lazy_query



from typing import (
    Any, Protocol, Sequence, Optional,
    Tuple, Dict, Union,
)


Row = Tuple[Any, ...]
CursorParams = Union[Dict[str, Any], Row]
CursorDescription = Sequence[Tuple[str, int, int, int, int, bool]]


class Cursor(Protocol):
    rowcount: int
    description: CursorDescription

    def execute(
            self,
            operation: str,
            parameters: CursorParams,
    ) -> None:
        pass

    def executemany(
            self,
            operation: str,
            seq_of_parameters: Sequence[CursorParams],
    ) -> None:
        pass

    def close(self) -> None:
        pass

    def fetchone(self) -> Row:
        pass

    def fetchmany(self, size: Optional[int]) -> Sequence[Row]:
        pass

    def fetchall(self):
        pass


class Connection(Protocol):
    autocommit: bool

    def close(self) -> None:
        pass

    def commit(self) -> None:
        pass

    def rollback(self) -> None:
        pass

    def cursor(self) -> Cursor:
        pass



from types import TracebackType
from typing import Optional, Type

from .types import Connection


class Transaction:

    def __init__(self, conn: Connection):
        self.conn = conn

    def __enter__(self):
        self.return_autocommit_initial = self.conn.autocommit
        if self.conn.autocommit is True:
            self.conn.autocommit = False
        return self

    def __exit__(
            self,
            type_: Optional[Type[BaseException]],
            value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        if type_ is None:
            self.conn.commit()
        else:
            self.conn.rollback()
        if self.return_autocommit_initial:
            self.conn.autocommit = True

        return False



from typing import TypedDict, List

import pytest

from classic.db_tools import Engine, Entity, Append, Mapper, Value


class Status(TypedDict):
    id: int
    title: str


class Task(TypedDict):
    id: int
    name: str
    statuses: List[Status]


mapper = Mapper(
    task=Entity(Task, 'id', statuses=Append('Status')),
    status=Value(Status, True),
)


sql = '''
    SELECT
        tasks.id            AS Task__id,
        tasks.name          AS Task__name,
        task_status.id      AS Status__id,
        task_status.title   AS Status__title
    FROM tasks
    LEFT JOIN task_status ON task_status.task_id = tasks.id
    ORDER BY tasks.id, task_status.id 
'''


@pytest.fixture
def tasks(engine: Engine, ddl):
    engine.query_from('example/save_task.sql').executemany([
        {'name': 'First', 'value': ''},
        {'name': 'Second', 'value': ''},
        {'name': 'Third', 'value': ''},
        {'name': 'Four', 'value': ''},
    ])
    engine.query_from('example/save_task_statuses.sql').executemany([
        {'title': 'CREATED', 'task_id': 1},
        {'title': 'CREATED', 'task_id': 2},
        {'title': 'CREATED', 'task_id': 3},
        {'title': 'STARTED', 'task_id': 1},
        {'title': 'FINISHED', 'task_id': 1},
    ])
    yield


@pytest.mark.parametrize('static', (True, False))
def test_returning_with_rels__all(engine: Engine, ddl, tasks, static):
    query = engine.query(sql, static=static).map_to(Task, mapper=mapper)
    assert query.all() == [
        Task(id=1, name='First', statuses=[
            Status(id=1, title='CREATED'),
            Status(id=4, title='STARTED'),
            Status(id=5, title='FINISHED'),
        ]),
        Task(id=2, name='Second', statuses=[
            Status(id=2, title='CREATED'),
        ]),
        Task(id=3, name='Third', statuses=[
            Status(id=3, title='CREATED'),
        ]),
        Task(id=4, name='Four', statuses=[]),
    ]


@pytest.mark.parametrize('static', (True, False))
def test_returning_with_rels__all__empty(engine: Engine, ddl, tasks, static):
    assert engine.query('''
        SELECT
            1 AS Task__id,
            1 AS Task__name,
            1 AS Status__id,
            1 AS Status__title
        FROM tasks
        WHERE FALSE
    ''', static=static).map_to(
        Task, mapper=mapper,
    ).all() == []


@pytest.mark.parametrize('static', (True, False))
def test_returning_with_rels__one(engine: Engine, ddl, tasks, static):
    obj = engine.query(sql, static=static).map_to(Task, mapper=mapper).one()
    assert obj == Task(id=1, name='First', statuses=[
        Status(id=1, title='CREATED'),
        Status(id=4, title='STARTED'),
        Status(id=5, title='FINISHED'),
    ])



from classic.db_tools import Engine, Mapper, Entity

from .dto import Task


mapper_sources = '''def mapper_func(rows):
    task_map = {}

    def map_task(rows):
        if row[0] is None:
            return None
        task_id = row[0]
        task = task_map.get(task_id)
        if task is None:
            task = task_map[task_id] = Task(id=row[0])
        return task
    last_root = None
    for row in rows:
        root = map_task(row)
        if last_root is not root:
            if last_root is not None:
                yield last_root
            last_root = root
    if last_root is not None:
        yield last_root'''


def test__mapper__sources(engine: Engine):
    query = engine.query(
        'SELECT 1 AS task__id'
    ).map_to(Task, mapper=Mapper(task=Entity(Task, 'id')))
    assert query.sources() == mapper_sources
